import type { NextPage } from "next";
import dynamic from "next/dynamic";
import { useRouter } from "next/router";
import { useCallback, useMemo, useState } from "react";
import { ExpressionCard } from "../../components";
import { Page } from "../../components/Page";
import { useExpressionCategories } from "../../hooks/useExpressionCategories";
import { useExpressionFilterQueryIds } from "../../hooks/useExpressionFilterQueryIds";
import { useExpressionSetQueryId } from "../../hooks/useExpressionSetQueryId";
import { useExpressionsInSet } from "../../hooks/useExpressionsInSet";
import {
  assignExpressionToSet,
  IndexedCategory,
  IndexedExpression,
} from "../../model";
import { sample } from "../../util/array-utils";
import { PageWithError } from "../../views/PageWithError/PageWithError";

const PageTitle = "Flash Card App - Practice";

// Do random selection here so we don't keep flipping states with interaction
const ExpressionPracticePage: NextPage = () => {
  // Query info
  const expression_set_id = useExpressionSetQueryId();
  const filter_ids = useExpressionFilterQueryIds();

  // Filter out failed expressions and select random expression
  const expressions = useExpressionsInSet(expression_set_id).filter(
    (expression) => !filter_ids.includes(expression.id!)
  );
  const expression: IndexedExpression | undefined = useMemo(
    () => (expressions ? sample(expressions) : undefined),
    [expressions]
  );

  // Fetch categories that the expression relates to
  const categories = useExpressionCategories(expression?.id || 0);

  // Fallback views for expression set content not found
  if (!expression_set_id) {
    return (
      <PageWithError title={PageTitle} message="Expression set not found" />
    );
  }
  if (!expression) {
    return (
      <PageWithError
        title={PageTitle}
        message="No expressions left in this set"
      />
    );
  }

  return (
    <ExpressionCardPracticeView
      key={expression.id}
      expression={expression}
      categories={categories}
    />
  );
};

interface ExpressionCardPracticeViewProps {
  expression: IndexedExpression;
  categories: IndexedCategory[];
}

// Handle internal state here
function ExpressionCardPracticeView({
  expression,
  categories,
}: ExpressionCardPracticeViewProps) {
  const [revealed, setRevealed] = useState(false);
  return (
    <Page title={PageTitle}>
      <div className="page-with-bottom-navigation">
        <section className="padding-small scroll">
          <ExpressionCard
            prompt={expression.prompt}
            categories={categories.map((category) => category.name)}
            description={expression.description}
            show_description={revealed}
          />
        </section>
        <section className="navigation-bottom">
          {revealed ? (
            <>
              <DemoteExpressionButton expression_id={expression.id!} />
              <PromoteExpressionButton expression_id={expression.id!} />
            </>
          ) : (
            <button
              className="navigation-item bottom text-navigation grow"
              onClick={() => setRevealed(true)}
            >
              Reveal
            </button>
          )}
        </section>
      </div>
    </Page>
  );
}

interface ExpressionIdProps {
  expression_id: number;
}

function PromoteExpressionButton({ expression_id }: ExpressionIdProps) {
  const expression_set_id = useExpressionSetQueryId();
  const handleClick = useCallback(() => {
    assignExpressionToSet({
      expression_id,
      expression_set_id: expression_set_id + 1,
    });
  }, [expression_id, expression_set_id]);

  return (
    <button
      className="navigation-item bottom text-navigation grow"
      onClick={handleClick}
    >
      <span>Right</span>
    </button>
  );
}

function DemoteExpressionButton({ expression_id }: ExpressionIdProps) {
  const { query, pathname, push } = useRouter();
  const expression_set_id = useExpressionSetQueryId();
  const handleClick = useCallback(() => {
    if (expression_set_id === 1) {
      const filter_ids = query["filter-ids"]
        ? `${query["filter-ids"]} ${expression_id}`
        : `${expression_id}`;
      push({
        pathname,
        query: {
          ...query,
          "filter-ids": filter_ids,
        },
      });
    } else {
      assignExpressionToSet({
        expression_id,
        expression_set_id: Math.max(1, expression_set_id - 1),
      });
    }
  }, [expression_id, expression_set_id, pathname, push, query]);

  return (
    <button
      className="navigation-item bottom text-navigation grow"
      onClick={handleClick}
    >
      <span>Wrong</span>
    </button>
  );
}

export default dynamic(() => Promise.resolve(ExpressionPracticePage), {
  ssr: false,
});
